"""
数据发送方接口.
"""

__all__ = ["Client"]

import base64
import json
from dataclasses import dataclass, field

from typing_extensions import TYPE_CHECKING, Optional, Tuple, Union, Dict

from .logger_factory import LoggerFactory

if TYPE_CHECKING:
    from _typeshed import GenericPath
else:
    GenericPath = str

from . import error
from .config import ClientConfig
from .crypto import ClientSessionKey, FileMode, ResponseKey, PrivateKey
from .ra import attest_server
from .utils import RepeatTimer, weak_repeat_timer


@dataclass(eq=False)
class Client:
    """
    数据发送方接口.

    本类的所有方法是线程安全的.
    """

    config: ClientConfig

    session_key: Optional[ClientSessionKey] = field(init=False, default=None)

    attest_timer: Optional[RepeatTimer] = field(init=False, default=None)

    root_key_info: Dict[str, PrivateKey] = field(init=False, default_factory=dict)

    logger = LoggerFactory.get_local_instance("pcc_jsc_client")

    def __post_init__(self):
        if self.config.root_key_config:
            self._root_key_parse()

        if self.config.attest_interval is not None:
            self.attest_timer = timer = weak_repeat_timer(
                self.config.attest_interval, self._attest_server_no_raise
            )
            timer.start()

            self._attest_server_no_raise()

    def _attest_server_no_raise(self):
        self.logger.info("attest_server active")
        try:
            self.attest_server(None)  # TODO: pass token in timer
        except Exception:
            self.logger.exception("Failed to attest server")

    def attest_server(self, token: Optional[str] = None) -> bool:
        """
        对数据接收方进行远程证明.
        """
        max_try_count = 3
        count = 0

        while count < max_try_count:
            count += 1
            ret, pub_key = attest_server(token, self.config)
            if ret:
                self.logger.info(f"ra attest success, pub_key={pub_key}")
                self.session_key = ClientSessionKey.load(pub_key)
                return True
            else:
                self.logger.info("ra attest failed, try again...")

        return False

    def _get_key(self) -> ClientSessionKey:
        # 线程安全: 将 session_key 字段保存为本地变量, 避免访问两次
        key = self.session_key
        if key is None:
            raise error.KeyMissingError()
        return key

    def encrypt(self, plaintext: Union[str, bytes]) -> str:
        """
        加密字符串或二进制数据.

        Returns:
            信封加密数据.

        Raises:
            KeyMissingError: 没有可用的密钥. 应通过 `attest_server` 验证数据接收方, 并获取密钥.
        """
        message, _ = self.encrypt_with_response(plaintext)
        return message

    def encrypt_with_response(self, plaintext: Union[str, bytes]) -> Tuple[str, ResponseKey]:
        """
        加密字符串或二进制数据, 并获取用于解密响应的密钥.

        Returns:
            信封加密数据, 及用于解密响应的密钥.

        Raises:
            KeyMissingError: 没有可用的密钥. 应通过 `attest_server` 验证数据接收方, 并获取密钥.
        """
        message, response_key = self._get_key().encrypt_with_response(plaintext)
        return message.serialize(), response_key

    def encrypt_file(
        self,
        source_path: GenericPath,
        dest_path: GenericPath,
        mode: FileMode,
    ) -> str:
        """
        加密文件数据.

        Returns:
            信封加密元数据. 可伴随加密文件发送给数据接收方.
        """
        metadata, _ = self.encrypt_file_with_response(source_path, dest_path, mode)
        return metadata

    def encrypt_file_with_response(
        self,
        source_path: GenericPath,
        dest_path: GenericPath,
        mode: FileMode,
    ) -> Tuple[str, ResponseKey]:
        """
        加密文件数据, 并获取用于解密响应的密钥.

        Returns:
            信封加密元数据, 及用于解密响应的密钥. 其中元数据可伴随加密文件发送给数据接收方.
        """
        key = self._get_key()

        if mode not in ["b", "t"]:
            raise error.ParamError("mode", "Mode should be 'b' or 't'")

        with open(source_path, f"r{mode}") as source, open(dest_path, f"w{mode}") as dest:
            metadata, response_key = key.encrypt_stream_with_response(source, dest, mode)

        return metadata.serialize(), response_key

    def _root_key_parse(self):
        try:
            root_key_config = self.config.root_key_config
            if not root_key_config:
                return

            if type(root_key_config) == str:
                root_key_config = json.loads(root_key_config)

            for app_info, pri_key_path in root_key_config.items():
                with open(pri_key_path) as f:
                    private_pem = f.read()
                self.root_key_info[app_info] = PrivateKey.from_private_pem(private_pem)

            self.logger.info("root_key_parse success")
        except Exception as e:
            self.logger.exception("root_key_parse exception={}".format(e))
            raise error.ConfigError("RootKey conf", "RootPriKey conf error")

    def gen_sign(self, app_info: str, msg: Union[str, bytes]) -> str:
        """
        生成签名
        :param app_info: app_id+app_name
        :param msg: app_id+app_name
        :return:
        """
        try:
            if not app_info or not msg:
                self.logger.error(f"gen_sign error, app_info is NULL")
                raise error.SignatureError("RSA signing failed") from None

            if app_info not in self.root_key_info:
                self.logger.info(f"gen_sign error, can not find app_info in root key conf, app_info={app_info}")
                raise error.SignatureError(f"can not find app_info={app_info}") from None

            return base64.b64encode(self.root_key_info.get(app_info).sign(msg)).decode()

        except Exception as e:
            self.logger.exception("gen_sign exception={}".format(e))
            raise error.SignatureError("can not find app_info") from e
