"""
数据接收方接口.
"""

__all__ = ["Server"]

import base64
import hashlib
import json
import requests

from dataclasses import dataclass, field
from datetime import datetime, timedelta, timezone
from threading import Lock
from typing_extensions import TYPE_CHECKING, Dict, List, Optional, Tuple, Union

if TYPE_CHECKING:
    from _typeshed import GenericPath
else:
    GenericPath = str

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
from ..tks.client import TKSClient, TKSConfig, utc8_timestamp
from . import error
from .config import ServerConfig
from .crypto import FileMode, ResponseKey, ServerSessionKey, PublicKey
from .logger_factory import LoggerFactory
from .ra import (
    RaRequest,
    RaResponse,
    RaResponseMetadata,
    RaResponsePod,
    RaResponsePods,
    prepare_ra_request,
    validata_ra_request,
)
from .service import EpsClient, RasClient
from .utils import RepeatTimer, weak_repeat_timer


@dataclass(eq=False)
class Server:
    """
    数据接收方接口.

    本类的所有方法是线程安全的.
    """

    config: ServerConfig

    ras_client: RasClient = field(init=False, repr=False)

    eps_client: EpsClient = field(init=False, repr=False)

    refresh_timer: Optional[RepeatTimer] = field(init=False, default=None)

    session_key: ServerSessionKey = field(init=False, default=None)

    old_session_key: ServerSessionKey = field(init=False, default=None)

    session_key_lock: Lock = field(init=False, default_factory=Lock)
    """
    线程安全: 修改 session_key 应该加锁.
    """

    root_key_info: Dict[str, PublicKey] = field(init=False, default_factory=dict)

    logger = LoggerFactory.get_local_instance("pcc_jsc_server")

    def __post_init__(self):
        try:
            self.ras_client = RasClient(self.config)
            self.eps_client = EpsClient(self.config)

            if self.config.root_key_config:
                self._root_key_parse()

            if self.config.refresh_interval is not None:
                self.refresh_timer = timer = weak_repeat_timer(
                    self.config.refresh_interval, self._auto_update_key
                )
                timer.start()

            self._auto_load_key()
        except Exception as e:
            self.logger.exception(f"__post_init__ failed, e={e}")

    def _auto_update_key(self):
        """
        非TKS密钥暂不更新
        :return:
        """
        try:
            self.logger.info("_auto_update_key")
            if self.config.tks_key_id is not None:
                self.load_tks_key()
        except Exception as e:
            self.logger.exception(f"Auto load key failed, e={e}")

    def _auto_load_key(self):
        try:
            self.logger.info("start load_key")
            if self.config.tks_key_id is not None:
                self.load_tks_key()

            elif self.config.key_file is not None:
                self.load_file_key()
            else:
                self.logger.info(
                    "Both TKS key id and key file options are empty. call generate_key"
                )
                self.generate_key()

        except Exception as e:
            self.logger.exception(f"Auto load key failed, e={e}")

    def _add_key(self, private_pem: Union[str, bytes], tks_version: Optional[int]):
        """
        添加密钥对.
        """
        # 判断是否需要更新密钥
        if type(private_pem) == str:
            new_md5 = hashlib.md5(private_pem.encode()).hexdigest()
        else:
            new_md5 = hashlib.md5(private_pem).hexdigest()

        if self.session_key:
            old_md5 = self.session_key.pem_md5
        else:
            old_md5 = ""

        if old_md5 != new_md5:
            expire_at = datetime.now(timezone.utc) + timedelta(seconds=self.config.valid_time)
            session_key = ServerSessionKey.load(private_pem, expire_at, tks_version, new_md5)
            with self.session_key_lock:
                self.old_session_key = self.session_key
                self.session_key = session_key

    def _pcc_login(self) -> Dict[str, str]:
        """
        访问pcc_controller模块，获取cookie
        :return:
        """

        if self.config.pcc_login_url is None:
            raise error.ConfigError("pcc_login_url", "PCC login url is missing")
        if self.config.pcc_password is None:
            raise error.ConfigError("pcc_password", "PCC password is missing")

        ts = str(utc8_timestamp())
        req_body = {
            "Username": self.config.pcc_app_id,
            "Password": hashlib.md5(self.config.pcc_password.encode() + ts.encode()).hexdigest()
        }

        response = requests.post(
            self.config.pcc_login_url, headers={"TimeStamp": ts}, json=req_body
        )
        cookies = response.cookies
        jeddak_pcc_session = cookies.get("jeddak_pcc_session")
        assert jeddak_pcc_session is not None
        cookies = {"jeddak_pcc_session": jeddak_pcc_session}
        return cookies

    def _get_pairkey_from_pem(self, private_key: bytes) -> Optional[Dict[str, str]]:
        """
        从私钥pem中产生公私钥对 for peixuan
        :param private_key:
        :return:
        """
        private_key_pem = serialization.load_pem_private_key(private_key, password=None)
        if not isinstance(private_key_pem, RSAPrivateKey):
            self.logger.error(f"not RSAPrivateKey, private_key={private_key}")
            return None

        pub_key = private_key_pem.public_key()
        pub_key_str = pub_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        ).decode("utf-8")

        ret = dict()
        ret["pub_key"] = pub_key_str
        ret["pri_key"] = private_key.decode()

        return ret

    def load_tks_key(self):
        """
        从 TKS 获取或更新密钥对.
        若配置中 `tks_app_id`, `tks_ring_id`, 或 `tks_key_id` 为空, 则报错.
        """

        app_id, ring_id, key_id = (
            self.config.tks_app_id,
            self.config.tks_ring_id,
            self.config.tks_key_id,
        )
        if app_id is None or ring_id is None or key_id is None:
            raise error.ConfigError("tks_key_id", "TKS key options are not specified")

        pcc_conf = None
        top_conf = None
        if self.config.bytedance_pcc_info:
            pcc_conf = json.loads(self.config.bytedance_pcc_info)
        elif self.config.bytedance_top_info:
            top_conf = json.loads(self.config.bytedance_top_info)

        cfg = TKSConfig(
            addr=self.config.tks_url,
            enable_tls=True,
            pcc_config=pcc_conf,
            top_config=top_conf,
        )

        tks_client = TKSClient(app_id, cfg)

        if self.config.pcc_login_url:
            cookies = self._pcc_login()
            private_key = tks_client.get_key(
                ring_id=ring_id, key_id=key_id, need_evidence=False, cookies=cookies
            )
        else:
            private_key = tks_client.get_key(ring_id=ring_id, key_id=key_id, need_evidence=False)

        version = None

        pair_key_info = self._get_pairkey_from_pem(private_key)
        self.logger.info(f"get RSA key from TKS, info={pair_key_info}")

        # TODO:
        try:
            self._add_key(private_key, version)
        except error.ParamError as e:
            if e.param == "private_pem":
                raise error.ConfigError("tks_key_id", e.message) from e.__cause__
            else:
                raise

    def load_file_key(self, path: Optional[GenericPath] = None):
        """
        从文件读取 RSA 密钥.

        Args:
            path: 密钥文件路径. 默认为配置中的 `key_file`.
        """

        path = path or self.config.key_file
        if path is None:
            raise error.ParamError("path", "Key file is not specified")

        with open(path) as f:
            self.load_key(f.read())

    def load_key(self, private_key: Union[str, bytes]):
        """
        导入 RSA 密钥.

        Args:
            private_key: PEM 格式的 RSA 私钥.
        """
        self._add_key(private_key, tks_version=None)

    def generate_key(self):
        """
        随机生成密钥对.
        """
        expire_at = datetime.now(timezone.utc) + timedelta(seconds=self.config.valid_time)

        session_key = ServerSessionKey.generate(expire_at)
        # 修改时加锁
        with self.session_key_lock:
            self.old_session_key = self.session_key
            self.session_key = session_key

    def remove_expired_keys(self):
        """
        删除已过期的密钥对.
        """
        # 修改时加锁
        with self.session_key_lock:
            if self.session_key.is_expired():
                self.session_key = None

            if self.old_session_key.is_expired():
                self.old_session_key = None

    def _get_public_key(self) -> str:
        """
        获取已保存的最新的公钥.

        Raises:
            KeyMissingError: 没有可用的密钥. 应通过 `load_tks_key`, `load_key`, 或
                `generate_key` 添加密钥.
        """
        with self.session_key_lock:
            current_key = self.session_key

        if not current_key:
            raise error.KeyMissingError() from None

        return current_key.get_public_key()

    def handle_ra_request(self, request: RaRequest) -> RaResponse:
        """
        处理来自数据发送方的远程证明请求.
        应当在调用本函数前, 验证请求中 `token` 身份令牌的有效性.

        Returns:
            远程证明响应. 可作为 JSON 响应给数据发送方.
        Raises:
            NetworkError: 访问证明生成服务 (EPS) 或远程证明服务 (RAS) 时出现问题.
            KeyMissingError: 没有可用的密钥. 应通过 `load_tks_key`, `load_key`, 或
                `generate_key` 添加密钥.
        """
        validata_ra_request(request)

        # quote = self.eps_client.get_quote(request["nonce"])
        # token = self.ras_client.get_attestation_evaluation("TDX", quote)

        res_meta_data = RaResponseMetadata(Code=0, Message="success")
        res_result: RaResponsePods = {
            "pod_1": RaResponsePod(
                evidence="evidence",
                token="token",
                key_info={"pub_key_info": self._get_public_key()},
            ),
        }

        return RaResponse(ResponseMetadata=res_meta_data, ResponseResult=res_result)

    def decrypt(self, message: Union[str, bytes], app_info: str = None, sign: str = None) -> bytes:
        """
        解密二进制数据.

        Raises:
            DecryptionError: 加密消息不合法, 或者加密使用的密钥不在当前可用的密钥中.
            KeyMissingError: 没有可用的密钥. 应通过 `load_tks_key`, `load_key`, 或
                `generate_key` 添加密钥.
        """
        if app_info and sign:
            if not self.verify_sign(app_info, sign):
                raise error.SignatureError("verify sign failed")

        plaintext, _ = self.decrypt_with_response(message)
        return plaintext

    def decrypt_with_response(self, message: Union[str, bytes],
                              app_info: str = None, sign: str = None) -> Tuple[bytes, ResponseKey]:
        """
        解密二进制数据, 并获取用于加密响应的密钥.

        Raises:
            DecryptionError: 加密消息不合法, 或者加密使用的密钥不在当前可用的密钥中.
            KeyMissingError: 没有可用的密钥. 应通过 `load_tks_key`, `load_key`, 或
                `generate_key` 添加密钥.
        """
        last_error: Optional[Exception] = None

        if app_info and sign:
            if not self.verify_sign(app_info, sign):
                raise error.SignatureError("verify sign failed")

        with self.session_key_lock:
            current_key = self.session_key
            old_key = self.old_session_key

        for key in [current_key, old_key]:
            try:
                response_key = key.decrypt(message)
                return response_key
            except Exception as e:
                last_error = e
                continue

        if last_error:
            # 尝试更新一下密钥，因为上面的失败可能是本地密钥版本晚于客户端的密钥版本，等TKS支持密钥版本后可用密钥版本号来比较
            self._auto_load_key()
            try:
                response_key = self.session_key.decrypt(message)
                return response_key
            except Exception as e:
                last_error = e

        raise last_error or error.KeyMissingError()

    def decrypt_file(
        self,
        metadata: Union[str, bytes],
        source_path: GenericPath,
        dest_path: GenericPath,
        mode: FileMode,
        app_info: str = None,
        sign: str = None
    ) -> ResponseKey:
        """
        解密文件数据.

        Returns:
            用于加密响应的密钥.
        """
        if mode not in ["b", "t"]:
            raise error.ParamError("mode", "Mode should be 'b' or 't'")

        if app_info and sign:
            if not self.verify_sign(app_info, sign):
                raise error.SignatureError("verify sign failed")

        with self.session_key_lock:
            current_key = self.session_key
            old_key = self.old_session_key

        last_error: Optional[Exception] = None
        for key in [current_key, old_key]:
            try:
                with open(source_path, f"r{mode}") as source, open(dest_path, f"w{mode}") as dest:
                    response_key = key.decrypt_stream(metadata, source, dest, mode)
                    return response_key
            except Exception as e:
                last_error = e
                continue

        if last_error:
            # 尝试更新一下密钥，因为上面的失败可能是本地密钥版本晚于客户端的密钥版本，等TKS支持密钥版本后可用密钥版本号来比较
            self._auto_load_key()
            try:
                with open(source_path, f"r{mode}") as source, open(dest_path, f"w{mode}") as dest:
                    response_key = self.session_key.decrypt_stream(metadata, source, dest, mode)
                    return response_key
            except Exception as e:
                last_error = e

        raise last_error or error.KeyMissingError()

    def attest_client(self, token: Optional[str] = None) -> bool:
        """
        对客户端进行远程证明, 判断客户端是否可信.

        Returns:
            True 为可信, False 为不可信.
        """
        if not self.config.client_ra_url:
            return True

        headers: Dict[str, str] = {}
        if token:
            headers["Authorization"] = f"Bearer {token}"
        request = prepare_ra_request()  # FIXME:
        try:
            response = requests.post(self.config.client_ra_url, headers=headers, json=request)
        except Exception as e:
            self.logger.exception(f"Network error: service={self.config.client_ra_url}")
            raise error.NetworkError("RA", self.config.client_ra_url, endpoint=None) from e

        if response and response.status_code != 200:
            self.logger.error(f"Service error: service={response} code={response.status_code}")
            raise error.ServiceError(
                "RA", self.config.client_ra_url, None, f"code={response.status_code}"
            )

        try:
            response_json = response.json()
            # check response_json
            self.logger.info(f"check response_json={response_json}")
            return True
        except Exception as e:
            self.logger.exception(f"Response is not JSON: service={self.config.client_ra_url}")
            raise error.ServiceError("RA", self.config.client_ra_url, None, "not JSON") from e

    def _root_key_parse(self):
        try:
            root_key_config = self.config.root_key_config
            if not root_key_config:
                return

            if type(root_key_config) == str:
                root_key_config = json.loads(root_key_config)

            for app_info, pub_key_path in root_key_config.items():
                with open(pub_key_path) as f:
                    public_pem = f.read()
                self.root_key_info[app_info] = PublicKey.from_public_pem(public_pem)

            self.logger.info("root_key_parse success")
        except Exception as e:
            self.logger.exception("root_key_parse exception={}".format(e))
            raise error.ConfigError("RootKey conf", "RootPubKey conf error")

    def verify_sign(self, app_info: str, msg: Union[str, bytes], sign: Union[str, bytes]) -> bool:
        """
        签名验证
        :param app_info: app_id+app_name
        :param msg:
        :param sign:
        :return:
        """
        try:
            if not app_info or not msg or not sign:
                self.logger.info(f"verify_sign error, app_info={app_info}, sign={sign}")
                raise error.SignatureError("RSA signing failed") from None

            if app_info not in self.root_key_info:
                self.logger.info(f"verify_sign error, can not find app_info in root key conf, app_info={app_info}")
                raise error.SignatureError(f"can not find app_info={app_info}") from None

            if isinstance(sign, str):
                sign = base64.b64decode(sign.encode())

            return self.root_key_info.get(app_info).verify(sign, msg)

        except Exception as e:
            self.logger.exception("verify_sign exception={}".format(e))
            raise error.SignatureError("can not find app_info") from e
