__all__ = [
    "RepeatTimer",
    "WeakMethod",
    "weak_repeat_timer",
    "generate_bytedance_token",
    "random_str",
    "hmac_sha256",
    "hash_sha256",
    "norm_query",
    "request_bytedance_gateway"
]

import hashlib
import logging
import uuid
import weakref
import hmac
import json
import requests

from threading import Timer
from datetime import datetime, timezone
from urllib.parse import quote

from typing_extensions import (
    Any,
    Callable,
    Concatenate,
    Generic,
    Iterable,
    Mapping,
    Optional,
    ParamSpec,
    Protocol,
    TypeVar,
)

logger = logging.Logger(__name__)


class RepeatTimer(Timer):
    """
    定时重复执行函数.
    """

    def run(self):
        while not self.finished.wait(self.interval):
            self.function(*self.args, **self.kwargs)


T = TypeVar("T")
P = ParamSpec("P")
R = TypeVar("R")
R_co = TypeVar("R_co", covariant=True)


class MethodProtocol(Protocol[T, P, R]):
    __self__: T

    __func__: Callable[Concatenate[T, P], R]

    __call__: Callable[P, R]


class WeakMethod(Generic[T, P, R_co]):
    """
    对方法 (bound method) 的弱引用.
    """

    instance: weakref.ReferenceType[T]
    func: weakref.ReferenceType[Callable[Concatenate[T, P], R_co]]

    def __init__(self, method: Callable[P, R_co]):
        # Type checkers do not differentiate methods from callables, so
        # we have to ignore the casting error
        _method: MethodProtocol[T, P, R_co] = method  # pyright: ignore

        self.instance = weakref.ref(_method.__self__)
        self.func = weakref.ref(_method.__func__)

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> Optional[R_co]:
        instance = self.instance()
        func = self.func()

        if instance is None or func is None:
            logger.info("Weak method called on freed object")
            return

        return func(instance, *args, **kwargs)


def weak_repeat_timer(
    interval: float,
    method: Callable[P, None],
    args: Optional[Iterable[Any]] = None,
    kwargs: Optional[Mapping[str, Any]] = None,
) -> RepeatTimer:
    """
    定时重复执行对象上的方法. 在对象被垃圾回收时自动结束线程.
    """
    # Due to limitations of 'ParamSpec', we cannot annotate 'args' and 'kwargs' precisely
    timer = RepeatTimer(interval, WeakMethod(method), args, kwargs)
    weakref.finalize(method.__self__, timer.cancel)  # pyright: ignore

    return timer


def generate_bytedance_token(ak: str, sk: str, timestamp: str) -> str:
    """
    生成访问bytedance top网关的token
    :param ak:
    :param sk:
    :param timestamp:
    :return:
    """
    h = hashlib.sha256()
    h.update(sk.encode())
    sk = h.hexdigest()
    h = hashlib.sha256()
    h.update((ak + sk + str(timestamp)).encode())
    token = h.hexdigest()
    return token


def random_str(count=6, pre=None):
    tmp = str(uuid.uuid4())
    ret = "".join(tmp.split("-"))[-count:]
    if pre:
        return pre + ret
    else:
        return ret


def hmac_sha256(key: bytes, content: str):
    return hmac.new(key, content.encode("utf-8"), hashlib.sha256).digest()


def hash_sha256(content: str):
    return hashlib.sha256(content.encode("utf-8")).hexdigest()


def norm_query(params):
    query = ""
    for key in sorted(params.keys()):
        if isinstance(params[key], list):
            for k in params[key]:
                query = query + quote(key, safe="-_.~") + "=" + quote(k, safe="-_.~") + "&"
        else:
            query = query + quote(key, safe="-_.~") + "=" + quote(params[key], safe="-_.~") + "&"
    query = query[:-1]
    return query.replace("+", "%20")


def request_bytedance_gateway(url: str, body: dict, headers: dict, top_config: dict):
    top_ak = top_config.get("ak")
    top_sk = top_config.get("sk")
    top_service = top_config.get("service")
    top_region = top_config.get("region", "cn-beijing")
    top_method = top_config.get("method", "POST")
    top_action = top_config.get("action", "GenerateRasAttestation")
    top_version = top_config.get("version", "2024-12-24")

    if not top_ak or not top_sk or not top_service:
        return None

    credential = {
        "access_key_id": top_ak,
        "secret_access_key": top_sk,
        "service": top_service,
        "region": top_region,
    }

    query = {}
    date_now = datetime.now(timezone.utc)
    request_param = {
        "body": json.dumps(body, separators=(",", ":")),
        "host": url,
        "path": "/",
        "method": top_method,
        "content_type": "application/json",
        "date": date_now,
        "query": {"Action": top_action, "Version": top_version, **query},
    }

    x_date = request_param["date"].strftime("%Y%m%dT%H%M%SZ")
    short_x_date = x_date[:8]
    x_content_sha256 = hash_sha256(request_param["body"])
    sign_result = {
        "Host": request_param["host"],
        "X-Content-Sha256": x_content_sha256,
        "X-Date": x_date,
        "Content-Type": request_param["content_type"],
    }

    signed_headers_str = ";".join(["content-type", "host", "x-content-sha256", "x-date"])

    canonical_request_str = "\n".join(
        [
            request_param["method"].upper(),
            request_param["path"],
            norm_query(request_param["query"]),
            "\n".join(
                [
                    "content-type:" + request_param["content_type"],
                    "host:" + request_param["host"],
                    "x-content-sha256:" + x_content_sha256,
                    "x-date:" + x_date,
                    ]
            ),
            "",
            signed_headers_str,
            x_content_sha256,
        ]
    )

    hashed_canonical_request = hash_sha256(canonical_request_str)
    credential_scope = "/".join(
        [short_x_date, credential["region"], credential["service"], "request"]
    )
    string_to_sign = "\n".join(
        ["HMAC-SHA256", x_date, credential_scope, hashed_canonical_request]
    )

    logger.debug(
        f"ra request={canonical_request_str}, hash={hashed_canonical_request}, sign={string_to_sign}"
    )

    k_date = hmac_sha256(credential["secret_access_key"].encode("utf-8"), short_x_date)
    k_region = hmac_sha256(k_date, credential["region"])
    k_service = hmac_sha256(k_region, credential["service"])
    k_signing = hmac_sha256(k_service, "request")
    signature = hmac_sha256(k_signing, string_to_sign).hex()

    sign_result["Authorization"] = (
        "HMAC-SHA256 Credential={}, SignedHeaders={}, Signature={}".format(
            credential["access_key_id"] + "/" + credential_scope,
            signed_headers_str,
            signature,
            )
    )

    headers = {**headers, **sign_result}

    r = requests.request(
        method=top_method,
        url="https://{}{}".format(request_param["host"], request_param["path"]),
        headers=headers,
        params=request_param["query"],
        data=request_param["body"],
    )
    return r
