import base64
import hashlib
import os
import secrets
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from typing import Any, Dict, Optional

import requests

from . import exceptions as exp
from .attester import EPSAttester
from .crypto import aes, ec

__all__ = [
    "TKSClient",
    "TKSConfig",
]

PATH_ATTEST = "/api/tks/v1/security/attest"
PATH_RING_CREATE = "/api/tks/v1/ring/create"
PATH_KEY_EXPORT = "/api/tks/v1/key/export"
PATH_KEY_CREATE = "/api/tks/v1/key/create"
PATH_KEY_IMPORT = "/api/tks/v1/key/import"
PATH_POLICY_SET = "/api/tks/v1/policy/set"
PATH_KEY_LIST = "/api/tks/v1/key/list"


def utc8_timestamp() -> int:
    return int(datetime.now(timezone(timedelta(hours=8))).timestamp())


@dataclass
class TKSConfig:
    addr: str = "jeddakchain.bytedance.com"
    enable_tls: bool = False
    pcc_config: dict = None
    top_config: dict = None


class TKSClient:
    def __init__(self, app_id: str, config: TKSConfig) -> None:
        self._app_id = app_id
        self._config = config
        self.attester = EPSAttester()

    def _request_tks(self,
                     url: str,
                     body: Any,
                     headers: Optional[Dict[str, str]] = None,
                     cookies: Optional[Dict[str, str]] = None,) -> dict:
        try:
            if self._config.top_config:
                from bytedance.jeddak_secure_channel.utils import request_bytedance_gateway
                resp = request_bytedance_gateway(url, body, headers, self._config.top_config)
            else:
                resp = requests.post(url=url, json=body, headers=headers, cookies=cookies)
        except Exception as e:
            raise exp.TKSError("Network error") from e
        try:
            resp.raise_for_status()
            resp = resp.json()
        except requests.HTTPError as e:
            raise exp.TKSError(
                "Status '{}' from TKS, URL: '{}', Content: {}".format(resp.status_code, url, resp.text)
            ) from e
        except requests.RequestException as e:
            raise exp.TKSError("Bad connection to TKS: {}".format(e)) from e

        if "Result" not in resp:
            raise exp.TKSError("Unexpected response: {}".format(resp))

        return resp["Result"]

    def challenge_server(self, nonce: str) -> str:
        return self._do_server_ra(nonce)["Report"]

    def create_ring(self, ring_name: str, desc: str = "", **kwargs) -> dict:
        try:
            if self._config.top_config:
                url = self._config.addr
                self._config.top_config["action"] = "CreateTksRing"
            else:
                url = self._get_tks_url(PATH_RING_CREATE)

            result = self._request_tks(
                url=url,
                body={
                    "RingName": ring_name,
                    "Description": desc
                },
                headers=self._get_headers(),
                **kwargs
            )
            return result
        except requests.RequestException as e:
            raise exp.TKSError("create_ring to TKS: {}".format(e)) from e

    def create_key(
        self,
        ring_id: str,
        algo: str,
        key_name: str = "",
        desc: str = "",
        **kwargs,
    ) -> dict:
        key_meta = self._do_key_creation(
            ring_id, algo=algo, key_name=key_name, src="internal", desc=desc, **kwargs
        )
        return key_meta

    def import_key(
        self,
        ring_id: str,
        algo: str,
        key: bytes,
        key_name: str = "",
        desc: str = "",
        **kwargs,
    ) -> dict:
        # Create an empty key record
        key_meta = self._do_key_creation(
            ring_id, algo=algo, key_name=key_name, src="external", desc=desc, **kwargs
        )

        # Locally generate key pair
        sk_local, pk_local = ec.generate_key_pair()

        # Challenge server
        ra_result = self._do_server_ra(bi_auth=False)
        pk_server = base64.b64decode(ra_result["DHParam"].encode())

        # Encrypt key
        ek = ec.diffie_hellman_key_exchange(sk_local, pk_server)
        encrypted_key = aes.gcm_encrypt(ek, key)

        if self._config.top_config:
            url = self._config.addr
            self._config.top_config["action"] = "ImportTksKey"
        else:
            url = self._get_tks_url(PATH_KEY_IMPORT)

        _ = self._request_tks(
            url=url,
            body={
                "RingID": ring_id,
                "KeyID": key_meta["KeyID"],
                "Key": base64.b64encode(encrypted_key).decode(),
                "DHParam": base64.b64encode(pk_local).decode(),
            },
            headers=self._get_headers(),
            **kwargs,
        )

        return key_meta

    def get_key(self, ring_id: str, key_id: str, need_evidence: bool = True, **kwargs) -> bytes:
        ra_result = self._do_server_ra(bi_auth=True)
        # get pub_key from server
        pk_server = base64.b64decode(ra_result["DHParam"].encode())

        sk_local, pk_local = ec.generate_key_pair()
        req_body = {
            "AppID": self._app_id,
            "RingID": ring_id,
            "KeyID": key_id,
            "DHParam": base64.b64encode(pk_local).decode(),
            "ClientChall": ra_result["Challenge"],
            # "RAEvidence": {"TEEType": "coco", "Report": base64.b64encode(evidence).decode()},
        }

        if need_evidence:
            evidence = self.attester.get_evidence(ra_result["Challenge"]["NonceDown"][:64])
            req_body["RAEvidence"] = {
                "TEEType": "coco",
                "Report": base64.b64encode(evidence).decode(),
            }

        if self._config.top_config:
            url = self._config.addr
            self._config.top_config["action"] = "ExportTksKey"
        else:
            url = self._get_tks_url(PATH_KEY_EXPORT)

        res = self._request_tks(url=url, body=req_body, headers=self._get_headers(), **kwargs)

        encrypted_key = res["Key"]

        ek = ec.diffie_hellman_key_exchange(local_sk=sk_local, peer_pk=pk_server)

        data_key = aes.gcm_decrypt(ek, base64.b64decode(encrypted_key))

        return data_key

    def set_key_policy(self, ring_id: str, key_id: str, policy: str, **kwargs) -> dict:
        if self._config.top_config:
            url = self._config.addr
            self._config.top_config["action"] = "ImportTksKey"
        else:
            url = self._get_tks_url(PATH_POLICY_SET)

        result = self._request_tks(
            url=url,
            headers=self._get_headers(),
            body={
                "ID": key_id,
                "Range": "key",
                "Rules": policy,
            },
            **kwargs,
        )
        return result

    def _do_key_creation(
        self,
        ring_id: str,
        algo: str,
        key_name: str = "",
        src: str = "internal",
        desc: str = "",
        **kwargs,
    ) -> dict:
        if self._config.top_config:
            url = self._config.addr
            self._config.top_config["action"] = "CreateTksKey"
        else:
            url = self._get_tks_url(PATH_KEY_CREATE)

        result = self._request_tks(
            url=url,
            body={
                "RingID": ring_id,
                "KeyName": key_name,
                "Algo": algo,
                "Source": src,
                "Description": desc,
            },
            headers=self._get_headers(),
            **kwargs,
        )

        return result

    def list_key(self, ring_id: str = None, page_number: int = 0, page_size: int = 10, status: str = None, **kwargs):
        """
        :param ring_id:
        :param page_size:
        :param page_number:
        :param status:
        :param kwargs:
        :return:
        """
        try:
            if self._config.top_config:
                url = self._config.addr
                self._config.top_config["action"] = "ListTksKey"
            else:
                url = self._get_tks_url(PATH_KEY_LIST)

            body = {"PageSize": page_size, "PageNumber": page_number}
            if ring_id:
                body["RingID"] = ring_id
            if status:
                body["Status"] = status

            result = self._request_tks(url=url, body=body, headers=self._get_headers(), **kwargs)
            return result
        except requests.RequestException as e:
            raise exp.TKSError("list key error: {}".format(e)) from e

    def _do_server_ra(self, nonce: Optional[str] = None, bi_auth: bool = True, **kwargs) -> dict:
        if nonce is None:
            nonce = secrets.token_hex(32)

        if self._config.top_config:
            url = self._config.addr
            self._config.top_config["action"] = "AttestTksSecurity"
        else:
            url = self._get_tks_url(PATH_ATTEST)

        resp = self._request_tks(
            url=url,
            body={"NonceUp": nonce, "BiAuth": bi_auth},
            headers=self._get_headers(),
            **kwargs)

        return resp

    def _get_tks_url(self, path: str) -> str:
        scheme = "https" if self._config.enable_tls else "http"
        return "{}://{}{}".format(scheme, self._config.addr, path)

    def _generate_pcc_token(self, timestamp: str) -> str:
        """
        生成访问bytedance pcc_controller的token
        :param timestamp:
        :return:
        """
        ak = self._config.pcc_config.get("ak")
        sk = self._config.pcc_config.get("sk")

        h = hashlib.sha256()
        h.update(sk.encode())
        sk = h.hexdigest()
        h = hashlib.sha256()
        h.update((ak + sk + str(timestamp)).encode())
        token = h.hexdigest()
        return token

    def _get_headers(self) -> dict:
        timestamp = str(int(datetime.now(timezone.utc).timestamp()))
        headers = {
            "AppID": self._app_id,
            # "Timestamp": str(utc8_timestamp()),
            "Timestamp": timestamp,
            # "nonce": "a314b43502077cd89d21378b29954207aaf5b3848d4d7fa9f16682ba1f6f3317"
        }

        if self._config.pcc_config:
            headers["Content-Type"] = "application/json"
            headers["J-Internal-Ak"] = self._config.pcc_config.get("ak")
            headers["J-Internal-Token"] = self._generate_pcc_token(timestamp)

        return headers

    def aes_encrypt(self, aes_key: [bytes, str], data: [bytes, str]) -> bytes:
        if type(aes_key) == str:
            aes_key = aes_key.encode()
        if type(data) == str:
            data = data.encode()
        return aes.gcm_encrypt(aes_key, data)

    def aes_decrypt(self, aes_key: [bytes, str], data: [bytes, str]) -> bytes:
        if type(aes_key) == str:
            aes_key = aes_key.encode()
        if type(data) == str:
            data = data.encode()
        return aes.gcm_decrypt(aes_key, data)


default_config = TKSConfig()


if __name__ == "__main__":

    def md5_hash(data: bytes, salt: str = "") -> str:
        return hashlib.md5(data + salt.encode()).hexdigest()

    ts = str(utc8_timestamp())
    req_body = {"Username": "alice", "Password": md5_hash("alice0".encode(), ts)}
    response = requests.post(
        "https://jeddakchain.bytedance.com/api/user/login", headers={"TimeStamp": ts}, json=req_body
    )
    cookies = response.cookies
    jeddak_pcc_session = cookies.get("jeddak_pcc_session")

    cookies = {"jeddak_pcc_session": jeddak_pcc_session}

    cfg = TKSConfig(addr="jeddakchain.bytedance.com", enable_tls=True)
    client = TKSClient("alice", cfg)
    # print(
    #     "Key Get",
    #     client.get_key(
    #         "2ee6348f-fafa-4e97-a824-0b6ed0e4d07e", "f532b84c-8f80-4705-b5a7-6217f4729924"
    #     ),
    # )
    #
    # print("Challenge Server", client.challenge_server("1234"))
    # print(
    #     "Key Creation",
    #     client.create_key(
    #         "4894a298-759a-4b94-a292-f43b2f8deea6",
    #         "SYMMETRIC_256",
    #         "test-encrypt-sdk",
    #         desc="测试本地加密SDK",
    #         cookies=cookies,
    #     ),
    # )
    #                                         )
    print(
        "Key Insert",
        client.import_key(
            "4894a298-759a-4b94-a292-f43b2f8deea6",
            "SYMMETRIC_256",
            os.urandom(32),
            "test-py-import",
            cookies=cookies,
        ),
    )
    # print("Key Set Policy", client.set_key_policy(
    #     "b3bf3cd7-2a68-427f-865f-8688dc96c7d8",
    #     "1110fb2f-8dca-4e27-9a37-04a60585fd29",
    #     json.dumps([
    #         [
    #             {
    #                 "version": "v1",
    #                 "type": "RAPolicy",
    #                 "optional": False,
    #                 "inversion": False,
    #                 "policyID": "78485ba8-4e67-46a5-a34e-1a9b7930df02"
    #             }
    #         ]
    #     ])
    # ))
